/*
 * Implementation of the interprocess communication
 *
 * HISTORY
 * Date         Author  Remarks
 * 1999-02-21   hjb     Created from code in other files
 */

#ifdef HAVE_MPI
#include <mpi.h>
#endif
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>

#include "netchaos.h"
#include "semlib.h"
#include "util.h"

static const int SIGHANDSHAKE = SIGUSR1;

#ifdef HAVE_MPI     
static int             recvFlag;
static MPI_Status      recvStatus;
static MPI_Request     request;
#endif

static int      resultSize0;
static DOUBLE*  result0 = NULL;

static enum
{
    COMM_SHM,
    COMM_MPI
} mode = COMM_MPI;

static SharedInfo* command;

static int          shmid;
static int          slaveSem;
static int          masterSem;


void CommSelect(int argc, char** argv)
{
#ifdef HAVE_MPI
    if (argc > 0)
    {
        if (strncmp(&argv[0][strlen(argv[0]) - 12], "netchaos-shm", 12) == 0)
            mode = COMM_SHM;
    }
#else
    mode = COMM_SHM;
#endif
}


void CommInitMaster(int argc, char** argv, int& nservers, int& rank)
{
#ifdef HAVE_MPI
    DOUBLE dummy;
#endif

    Debug(DEBUG_COMM(3), "CommInitMaster entered\n");

    switch (mode)
    {
        case COMM_SHM:
            CreateChild(command);
            if ((slaveSem = sem_create(MY_SHM_KEY + 1, 1)) < 0)
            {
                Debug(0, "Can't create semaphore\n");
                Exit(1);
            }
            if ((masterSem = sem_create(MY_SHM_KEY + 2, 0)) < 0)
            {
                Debug(0, "Can't create semaphore\n");
                Exit(1);
            }
            nservers = 2;
        break;
#ifdef HAVE_MPI
        case COMM_MPI:
            command = (SharedInfo *) xmalloc(sizeof(*command));
            MPI_Init(&argc, &argv);

            MPI_Comm_size(MPI_COMM_WORLD, &nservers);
            if (nservers < 2)
            {
                Debug(0, "No servers available\n");
                Exit(1);
            }

            MPI_Comm_rank(MPI_COMM_WORLD, &rank);
            if (rank != 0)
            {
                Calc();
                // Never reached.
                Debug(DEBUG_GEN(0), "Reached anyway\n");
            }
            MPI_Irecv(&dummy, 1, MPI_DOUBLE, MPI_ANY_SOURCE,
                0, MPI_COMM_WORLD, &request);
        break;

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommInitMaster\n");
    }
#endif
}


void CommRecvCancel()
{
    Debug(DEBUG_COMM(3), "CommRecvCancel entered\n");

    switch (mode)
    {
        case COMM_SHM:
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Cancel(&request);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvCancel\n");
    }
}


void CommRecvNew()
{
    Debug(DEBUG_COMM(3), "CommRecvNew entered\n");

    switch (mode)
    {
        case COMM_SHM:
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Irecv(result0, resultSize0, MPI_DOUBLE, MPI_ANY_SOURCE,
                0, MPI_COMM_WORLD, &request);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvNew\n");
    }
}


void CommRecvNext()
{
    Debug(DEBUG_COMM(3), "CommRecvNext entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_wait(masterSem);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            // Or CommRecvNew()
            MPI_Irecv(result0, resultSize0, MPI_DOUBLE, MPI_ANY_SOURCE,
                0, MPI_COMM_WORLD, &request);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvNext\n");
    }
}


Boolean CommRecvTestMaster()
{
    Debug(DEBUG_COMM(255), "CommRecvTest entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_wait(masterSem);
            return True;
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Test(&request, &recvFlag, &recvStatus);
            return recvFlag != 0;
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvTest\n");
    }

    return False;
}


Boolean CommRecvTestSlave(int imageno)
{
    Debug(DEBUG_COMM(255), "CommRecvTestSlave entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_wait(slaveSem);
            return True;
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Test(&request, &recvFlag, &recvStatus);
            return recvFlag != 0 && command->imageno != imageno;
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvTestSlave\n");
    }

    return False;
}


int CommRecvSource()
{
    switch (mode)
    {
        case COMM_SHM:
            return 1;
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            return recvStatus.MPI_SOURCE;
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvSource\n");
    }

    return 1;
}


void CommSend(int server)
{
    // Called by master (which is the client!)
    Debug(DEBUG_COMM(3), "CommSend entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_signal(slaveSem);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Send(command, sizeof(*command), MPI_BYTE, server,
                0, MPI_COMM_WORLD);
            break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommSend\n");
    }
}


Boolean CommInitSlave()
{
    Debug(DEBUG_COMM(3), "CommInitSlave entered\n");

    switch (mode)
    {
        case COMM_SHM:
            // Allocate only the known size of SharedInfo; the rest
            // is allocated later.
            shmid = shmget(MY_SHM_KEY, sizeof(SharedInfo), 0);
            if (shmid < 0)
            {
                Debug(DEBUG_GEN(0), "Could not get shared memory\n");
                return False;
            }

            command = (SharedInfo *) shmat(shmid, 0, 0);
            if ((int) command < 0)
            {
                Debug(DEBUG_GEN(0), "Could not attach shared memory\n");
                return False;
            }

        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Irecv(command, sizeof(*command), MPI_BYTE, 0, 0,
                MPI_COMM_WORLD, &request);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommInitSlave\n");
            return False;
    }

    return True;
}


void CommWaitSlave()
{
    Debug(DEBUG_COMM(3), "CommWait entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_wait(slaveSem);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Wait(&request, &recvStatus);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommWait\n");
    }
}


void CommRecvNextSlave()
{
    Debug(DEBUG_COMM(3), "CommRecvNextSlave entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_signal(masterSem);
            sem_wait(slaveSem);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Irecv(command, sizeof(*command), MPI_BYTE, 0, 0,
                MPI_COMM_WORLD, &request);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommRecvNextSlave\n");
    }
}


int CommGetResultSize(Image *pic)
{
    resultSize0 = HEADER_SIZE;

    switch (mode)
    {
        case COMM_SHM:
            resultSize0 += pic->width;
        break;

        case COMM_MPI:
            resultSize0 += (pic->height / MAX_LINE_FACTOR) * pic->width;
        break;

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommGetResultSize\n");
    }

    return resultSize0;
}


SharedInfo *CommGetCommand()
{
    return command;
}


Boolean CommAllocBuffer(int size)
{
    Debug(DEBUG_COMM(3), "CommAllocBuffer entered\n");

    resultSize0 = size;

    switch (mode)
    {
        case COMM_SHM:
            shmdt((char *) command);

            /* Now we open again with the correct length */
            shmid = shmget(MY_SHM_KEY, sizeof(*command)
                + sizeof(DOUBLE) * size, 0600);
            if (shmid < 0)
            {
                Debug(DEBUG_GEN(0), "Could not get shared memory\n");
                return False;
            }

            command = (SharedInfo *) shmat(shmid, 0, 0);
            if ((int) command < 0)
            {
                Debug(DEBUG_GEN(0), "Could not get attach shared memory\n");
                return False;
            }

            result0 = (DOUBLE *) &command[1];
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            result0 = (DOUBLE *) xrealloc(result0, sizeof(DOUBLE) * size);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommAllocBuffer\n");
    }

    return True;
}


DOUBLE *CommGetBuffer()
{
    return result0;
}


void CommFreeBuffer()
{
    xfree(result0);
}


void CommSendResponse(DOUBLE response)
{
    Debug(DEBUG_COMM(3), "CommSendResponse entered\n");

    switch (mode)
    {
#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Send(&response, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommSendResponse\n");
    }
}


void CommSendResult()
{
    Debug(DEBUG_COMM(3), "CommSendResult entered\n");

    switch (mode)
    {
        case COMM_SHM:
            sem_signal(masterSem);
            sem_wait(slaveSem);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Send(result0, resultSize0, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommSendResult\n");
    }
}


void CommEndMaster()
{
    Debug(DEBUG_COMM(3), "CommEndMaster entered\n");

    switch (mode)
    {
        case COMM_SHM:
            Shutdown(command);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            //command->cmd = CMD_QUIT;
            //MPI_Send(command, sizeof(*command), MPI_BYTE, 1, 0, MPI_COMM_WORLD);
            MPI_Finalize();
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommEndMaster\n");
    }
}


void CommEndSlave()
{
    Debug(DEBUG_COMM(3), "CommEndSlave entered\n");

    switch (mode)
    {
        case COMM_SHM:
            shmdt((char *) command);
        break;

#ifdef HAVE_MPI
        case COMM_MPI:
            MPI_Finalize();
            xfree(command);
        break;
#endif

        default:
            Debug(DEBUG_GEN(0), "Internal error in CommEndSlave\n");
    }
}
