/*
 * Implementation of the Mandelbrot calculation
 *
 *  HISTORY
 *  Date        Author  Remarks
 *  1996-10-17  hjb     Reformatted
 *  1998-02-10  hjb     New algorithm for count
 *  1999-02-16  hjb     Optimized assembly routine called
 */

#include <math.h>
#include "mdefs.h"
#include "netchaos.h"


extern "C" int get_cpu_type();
extern "C" double iter_pentium_II(Image *image);
extern "C" double iter_pentium_classic(Image *image);
extern "C" double iter_i486(Image *image);


typedef double (*IterFunction)(Image *);


LONG iterations;


static IterFunction iterFunc;
char x86_vendor[13];
int x86;
int x86_model;


void InitIteration()
{
#ifdef __i386__
    int cpu = get_cpu_type();
    Debug(1, "CPUID = %.8x\n", cpu);

    iterFunc = iter_i486;

    if (x86 < 5)
        return;

    if (strncmp(x86_vendor, "GenuineIntel", 12) != 0)
        return;

    x86 = (cpu & 0x00000f00) >> 8;
    x86_model = (cpu & 0x000000f0) >> 4;

    if (x86 == 5)
        iterFunc = iter_pentium_classic;
    else
        iterFunc = iter_pentium_II;
#endif
}


double Iterate(Image *image)
{
#ifdef __i386__
    return (*iterFunc)(image);
#else

    volatile double  modulus;
    volatile DOUBLE  qre, qim;
    volatile DOUBLE  zre, zim;
    volatile DOUBLE  cre, cim;
    volatile DOUBLE  bailout = image->bailout;
    volatile LONG    count = 0;
    volatile LONG    maxc = image->maxc;

    zre = image->reC;
    zim = image->imC;

    if (image->julia)
    {
        cre = image->reJ;
        cim = image->imJ;
    }
    else
    {
        cre = image->reC;
        cim = image->imC;
    }

    cre += image->reP;
    cim += image->imP;

    qre = zre * zre;
    qim = zim * zim;

    do
    {
        count++;
        zim = 2 * zre * zim + cim;
        zre = qre - qim + cre;
        qre = zre * zre;
        qim = zim * zim;
    } while (count < maxc && qre + qim < bailout);
    //Debug(0, "qre = %f, qim = %f\n", qre, qim);

    if (count == maxc)
    {
        iterations += count;
        return (double) count;
    }

    count++;
    zim = 2 * zre * zim + cim;
    zre = qre - qim + cre;
    qre = zre * zre;
    qim = zim * zim;
    //Debug(0, "qre = %f, qim = %f\n", qre, qim);
    count++;
    zim = 2 * zre * zim + cim;
    zre = qre - qim + cre;
    qre = zre * zre;
    qim = zim * zim;
    //Debug(0, "qre = %f, qim = %f\n", qre, qim);

    modulus = sqrt(qre + qim);

    iterations += count;

    //Debug(0, "Count = %ld, mod = %f\n", count, modulus);
    return (double) count - (log(log(modulus)) / log(2.0));

#endif /* __i386__ */

} /* Iterate */

#ifdef _TEST
#include <stdio.h>
Boolean sharedmem;

int main()
{
    struct Image    image;
    double          d;

    image.re1 = -2.5;
    image.re2 = 1.5;
    image.im1 = 2.0;
    image.im2 = -2.0;
    image.reC = -2.5;
    image.imC = 1.5;
    image.reJ = 0.0;
    image.imJ = 0.0;
    image.reP = 0.0;
    image.imP = 0.0;
    image.bailout = 4.0;
    image.maxc = 64000;
    image.julia = 0;

    InitIteration();
    d = Iterate(&image);
    printf("d = %.4f\n", d);
}
#endif
