/*
 * Implementation of the main module
 *
 *  HISTORY
 *  Date        Author  Remarks
 *  1996-10-17  hjb     Reformatted
 *  1998-02-14  hjb     Assume 'Iterate' returns a double
 *  1999-02-04  hjb     Fix definitions and includes for libc6
 *  1999-02-10  hjb     Send multiple lines at once
 *  1999-02-11  hjb     Calculate number of lines based on processor speed
 *  1999-02-14  hjb     Shared mem version (again :)
 *  1999-02-19  hjb     Command line option -d; HAVE_MPI
 *  1999-02-21  hjb     Moved all MPI/sharedmem dependencies into another file
 *                      Moved xalloc etc. to util.c
 */

#include <getopt.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw3d/Reports.h>
#include <X11/Xaw3d/Simple.h>
#include <X11/Xaw3d/Viewport.h>

#include "netchaos.h"
#include "comm.h"
#include "mdefs.h"          /* portability */
#include "dialog.h"
#include "EditField.h"
#include "util.h"

#ifndef PI
#define PI M_PI
#endif


/* Titles of the main window */

#define PROGNAME "NetChaos"
#define PROGCALC "NetChaos (calculating)"

typedef struct
{
    int     line;               /* current line; -1 if idle */
    int     nlines;             /* outstanding lines */
    DOUBLE  speed;              /* iterations/s */
} ServerInfo;

typedef struct
{
    int         serverno;       /* not used yet */
    ServerInfo* server;         /* not used yet */
    DOUBLE      iterations;     /* 0 if line not yet calculated */
} LineInfo;

Image       pic;
ScreenInfo  mscreen;
GC          gc;

Rubberband  rubberband;

struct
{
  Image     *stack;
  int       level;
} stack;

Boolean     newPicture;





/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

void Exit(int status)
{
    CommEndMaster();
    exit(status);

} /* Exit */




/*---------------------------------------------------------------------------
 *
 * DANG_BEGIN_FUNCTION SaveToStack
 * Saves the current image to a stack where it can be retrieved later. 
 * DANG_END_FUNCTION SaveToStack
 *
 *---------------------------------------------------------------------------*/

void SaveToStack()
{
    Image *tempstack;

    ++stack.level;
    tempstack = (Image *) realloc(stack.stack, stack.level * sizeof(Image));
    if (tempstack == 0)
    {
        Debug(DEBUG_GEN(0), "Out of memory\n");
        Exit(1);
    }

    Debug(1, "Image saved to stack, level %d\n", stack.level);

    stack.stack = tempstack;
    memcpy(&stack.stack[stack.level - 1], &pic, sizeof pic);

} /* SaveToStack */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

void RestoreFromStack()
{
    if (stack.level > 1)
    {
        stack.level--;
        memcpy(&pic, &stack.stack[stack.level], sizeof pic);
        /* FIXME look whether we can load the picture from file */
        newPicture = True;
        pic.row = 0;
        Debug(1, "Image fetched from stack. New level %d\n", stack.level);
    }

} /* RestoreFromStack */




/*---------------------------------------------------------------------------
 *
 * Depth 8 only.
 *
 *---------------------------------------------------------------------------*/

void GetColormap(XColor *colors)
{
    int i;

    for (i = 0; i < 256; i++)
        colors[i].pixel = i;

    XQueryColors(XtDisplay(mscreen.toplevel), mscreen.colormap, colors, 256);

} /* GetColormap */




/*---------------------------------------------------------------------------
 *
 * Sets a colormap. For depth 8 only.
 *
 *---------------------------------------------------------------------------*/

void SetColormap(Widget w)
{
#if 1
    XColor    color[256];
    int       i;
    double    theta;
    double    saturation;
    int       intensity;

    mscreen.colormap = XCreateColormap(XtDisplay(w), XtWindow(w),
        DefaultVisualOfScreen(XtScreen(w)), AllocAll);
    /* Here user defined colormaps could be loaded */
    color[0].pixel = 0;
    color[0].red = 0;
    color[0].green = 0;
    color[0].blue = 0;
    color[0].flags = DoRed | DoGreen | DoBlue;
    for (i = 1; i < 256; i++)
    {
        /* Take the count as a HSI value, with H changing fastest.
        Say I 0 bit, S 1 bit, H 6 bit. */
        theta = 2 * PI * ((256 + 80 - i) & 127) / 128.0;
        saturation = ((((256 - i) >> 7) & 1) + 1) / 2.0;
        intensity = 2;
        color[i].pixel = i;
        color[i].red = (int) floor(intensity * 16383.0 * (1 + saturation * sin(theta - 2 * PI / 3)));
        color[i].green = (int) floor(intensity * 16383.0 * (1 + saturation * sin(theta)));
        color[i].blue = (int) floor(intensity * 16383.0 * (1 + saturation * sin(theta + 2 * PI / 3)));
        color[i].flags = DoRed | DoGreen | DoBlue;
    }
    XStoreColors(XtDisplay(w), mscreen.colormap, color, 256);
    XSetWindowColormap(XtDisplay(w), XtWindow(w), mscreen.colormap);
#else
    XColor    color;
    int       i;

    mscreen.colormap = DefaultColormapOfScreen(XtScreen(w));
    /* Here user defined colormaps could be loaded */
    for (i = 0; i < 256; i++)
    {
        color.pixel = i;
        color.red = (255 - i) << 8;
        color.green = (255 - i) << 8;
        color.blue = (255 - i) << 8;
        color.flags = DoRed | DoGreen | DoBlue;
        XAllocColor(XtDisplay(w), mscreen.colormap, &color);
    }
#endif
} /* SetColormap */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

int PixelColor(double count, int depth)
{
    int       r, g, b;
    double    theta;
    double    saturation;
    int       intensity;
    unsigned  temp;

    if (count >= (double) pic.maxc)
        return 0;
    else
        switch (depth)
        {
        case 8:
            return 255 - ((unsigned) (count * 128)) % 255;
        break;
  
        case 15:
            /* color weight 5-5-5 */
            /* Take the count as a HSI value, with H changing fastest.
                Say I 4 bit, S 5 bit, H 6 bit. */
            temp = (unsigned) (count * 128);
            temp = 0x8000 - (temp & 0x7fff);
            /* The offset to temp 'rotates' the colors. */
            theta = 2 * PI * ((temp + 40) & 63) / 64.0;
            saturation = (((temp >> 6) & 31) + 1) / 32.0;
            intensity = (temp >> 11) & 15;
            r = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta - 2 * PI / 3)));
            g = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta)));
            b = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta + 2 * PI / 3)));
            return (r << 10) + (g << 5) + b;
        break;
  
        case 16:
            /* Take the count as a HSI value, with H changing fastest.
            Say I 5 bit, S 5 bit, H 6 bit. */
            temp = (unsigned) (count * 128);
            temp = 0x10000 - (temp & 0xffff);
            /* The temp to count 'rotates' the colors. */
            theta = 2 * PI * ((temp + 40) & 63) / 64.0;
            saturation = (((temp >> 6) & 31) + 1) / 32.0;
            intensity = (temp >> 11) & 31;
            if (mscreen.weight565)
            {
                r = (int) floor(intensity * 0.5 * (1 + saturation * sin(theta - 2 * PI / 3)));
                g = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta)));
                b = (int) floor(intensity * 0.5 * (1 + saturation * sin(theta + 2 * PI / 3)));
                return (r << 11) + (g << 5) + b;
            }
            else
            {
                r = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta - 2 * PI / 3)));
                g = (int) floor(intensity * 0.5 * (1 + saturation * sin(theta)));
                b = (int) floor(intensity * 0.5 * (1 + saturation * sin(theta + 2 * PI / 3)));
                return (r << 10) + (g << 5) + b;
            }
        break;
  
      case 24:
            /* color weight 8-8-8 */
            /* Take the count as a HSI value, with H changing fastest.
                Say I 7 bit, S 8 bit, H 9 bit. */
            temp = (unsigned) (count * 128);
            temp = 0x1000000 - (temp & 0xffffff);
            /* The offset to temp 'rotates' the colors. */
            theta = 2 * PI * ((temp + 320) & 511) / 512.0;
            saturation = (((temp >> 9) & 255) + 1) / 256.0;
            intensity = (temp >> 17) & 127;
            r = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta - 2 * PI / 3)));
            g = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta)));
            b = (int) floor(intensity * 1.0 * (1 + saturation * sin(theta + 2 * PI / 3)));
            return (r << 16) + (g << 8) + b;
        break;
  
        default:
            Debug(DEBUG_GEN(0), "Don't know how to handle depth %d\n", depth);
            Exit(1);
    } /* switch */

    return 0;            /* Never reached */

} /* PixelColor */




/*---------------------------------------------------------------------------
 *
 * Depth 16 only.
 *
 *---------------------------------------------------------------------------*/

void DetectColorWeight(Display *display, Screen *screen)
{
    XColor    color;

    color.pixel = 31 << 5;
    XQueryColor(display, DefaultColormapOfScreen(screen), &color);
    mscreen.weight565 = (color.green < 32768);

} /* DetectColorWeight */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

void EnterWindowCB(Widget w, XtPointer clientData, XEvent *event,
    Boolean *cont)
{
    XSetWindowColormap(XtDisplay(w), XtWindow(w), mscreen.colormap);
} /* EnterWindowCB */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

void InitImage(Image *pic)
{
    pic->re1 = -2.5;
    pic->re2 = 1.5;
    pic->im1 = 1.5;
    pic->im2 = -1.5;
    pic->reP = 0.0;
    pic->imP = 0.0;
    pic->reJ = (pic->re2 + pic->re1) / 2;
    pic->imJ = (pic->im2 + pic->im1) / 2;
    pic->bailout = 4.0;
    pic->width = 240;
    pic->height = 180;
    pic->maxc = 100;
    pic->julia = False;
}




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

void InitEventHandlers()
{
    XtAddEventHandler(mscreen.toplevel, EnterWindowMask, False,
        EnterWindowCB, 0);
    XtAddEventHandler(mscreen.draw, KeyPressMask, False, KeyPressCB, 0);
    XtAddEventHandler(mscreen.draw, ButtonPressMask, False, ButtonPressCB, 0);
    XtAddEventHandler(mscreen.draw, ButtonReleaseMask, False, ButtonReleaseCB, 0);
    XtAddEventHandler(mscreen.draw, PointerMotionMask, False, PointerMotionCB, 0);
}




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

ServerInfo *InitServers(int nservers)
{
    int i;
    ServerInfo *server = (ServerInfo *) xmalloc(sizeof(ServerInfo) * nservers);

    for (i = 0; i < nservers; i++)
    {
        server[i].line = -1;
        server[i].nlines = 0;
        server[i].speed = 0.0;
    }

    return server;
}




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

void DisplayLine(DOUBLE* result, int line, int width)
{
    int column;

    for (column = 0; column < width; column++)
        XPutPixel(mscreen.image, column, line,
            PixelColor(result[column + HEADER_SIZE], mscreen.depth));

    if (rubberband.Visible())
    {
        /* delete rubberband, restore it afterwards */
        rubberband.Draw(mscreen.draw);
    }

    XPutImage(mscreen.display, XtWindow(mscreen.draw), gc, mscreen.image,
        0, line, 0, line, width, 1);
    if (rubberband.Visible())
    {
        rubberband.Draw(mscreen.draw);
    }
}




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    XtAppContext    appContext;
    Widget          canvas;
    Screen          *screen;
    int             nservers;
    ServerInfo      *server = NULL;
    int             freeservers = 0;
    int             nextserver = 0;
    int             rank;               /* 0 means master */
    int             resultSize = 0;
    DOUBLE          *result = NULL;
    DOUBLE          totalspeed = 0.0;
    LineInfo        *lines = NULL;
    SharedInfo      *command;
    int             imageno = 0;
    int             outstandingLines = 0;
    int             calculatedLines = 0;
    int             nextline = 0;
    struct timeval  tv;
    struct timeval  tvStart;

    enum
    {
        C_DISPATCH,
        C_INIT,
        C_TEST,
        C_LAUNCH
    } state = C_DISPATCH;

    struct option hOption[] =
    {
        { "debug", required_argument, NULL, 'd' },
        { NULL, 0, NULL, 0 }
    };

    int optind = 0;
    int optChar;


    CommSelect(argc, argv);         /* Select communication mode */
    InitIteration();                /* Select optimized assembler routine */
    InitImage(&pic);
    CommInitMaster(argc, argv, nservers, rank);

    mscreen.toplevel = XtOpenApplication(&appContext, PROGNAME, 0, 0,
        &argc, argv, 0, applicationShellWidgetClass, 0, 0);
    XtVaSetValues(mscreen.toplevel, XtNtitle, PROGNAME, 0);

    canvas = XtVaCreateManagedWidget("canvas", viewportWidgetClass,
        mscreen.toplevel, 0);
    mscreen.draw = XtVaCreateManagedWidget("draw", simpleWidgetClass, canvas,
        XtNwidth, pic.width, XtNheight, pic.height, 0);
    XtAddCallback(canvas, XtNreportCallback, ResizeCB, (XtPointer) mscreen.draw);

    XtRealizeWidget(mscreen.toplevel);

    /* Create dialog boxes */
    AboutInit(mscreen.toplevel);
    FileSelectInit(mscreen.toplevel);
    MessageInit(mscreen.toplevel);
    OptionInit(mscreen.toplevel);
    QuitInit(mscreen.toplevel);
    SizeInit(mscreen.toplevel);

    mscreen.display = XtDisplay(canvas);
    screen = XtScreen(canvas);
    gc = DefaultGCOfScreen(screen);
    mscreen.depth = DefaultDepthOfScreen(screen);
    if (mscreen.depth < 8)
    { 
        Debug(DEBUG_GEN(0), "Depths smaller than 8 are not supported\n");
        Exit(1);
    }
    else if (mscreen.depth == 8)
    {
        SetColormap(mscreen.draw);
    }
    else if (mscreen.depth == 16)
    {
        DetectColorWeight(mscreen.display, screen);
    }

    mscreen.bpp = (mscreen.depth + 7) >> 3;    /* e.g. depth 15, bpp 2 */
    if (mscreen.bpp == 3)
        mscreen.bpp = 4;

    do
    {
        switch (optChar = getopt_long(argc, argv, "d", hOption, NULL))
        {
            case ':':
                // Missing argument.
                Debug(DEBUG_GEN(0), "Missing argument for option %c\n", optChar);
                break;
            case '?':
                // Invalid option.
                Debug(DEBUG_GEN(0), "Invalid option %c\n", optChar);
                break;
            case 'd':
                SetDebugLevel(atoi(optarg));
                break;
        }
    } while (optChar != EOF);

    signal(SIGTERM, Exit);
    signal(SIGINT, Exit);

    mscreen.image = 0;
    NewSize();
    rubberband.SetScreen(screen);
    stack.stack = 0;
    stack.level = 0;
    SaveToStack();
    InitEventHandlers();

    newPicture = True;
    pic.row = 0;

    while (True)
    {
        while (XtAppPending(appContext))
        {
            XEvent event;
            XtAppNextEvent(appContext, &event);
            XtDispatchEvent(&event);
        }

        switch (state)
        {
            case C_DISPATCH:
                if (newPicture)
                {
                    state = C_INIT;
                }
                else if (outstandingLines)
                {
                    state = C_TEST;
                }
                else // idle
                {
                    XEvent event;
                    XtAppNextEvent(appContext, &event);
                    XtDispatchEvent(&event);
                }
            break;

            case C_INIT:
                state = C_TEST;
                newPicture = False;
                GetTime(tvStart);

                /* This is better than calling ExposeCB because the new image
                    can be smaller than the old. */
                XClearWindow(mscreen.display, XtWindow(mscreen.draw));
                //XGetGCValues(mscreen.display, gc, GCBackground, &values);

                XtVaSetValues(mscreen.toplevel, XtNtitle, PROGCALC, 0);
                //for (int ro = 0; ro < pic.height; ro++)
                    //for (int co = 0; co < pic.width; co++)
                        //XPutPixel(mscreen.image, co, ro, values.background);
                //memset(mscreen.data, values.background, mscreen.bpp * pic.width * pic.height);
                ++imageno;

                resultSize = CommGetResultSize(&pic);

                CommAllocBuffer(resultSize);
                result = (DOUBLE *) CommGetBuffer();
                lines = (LineInfo *) xrealloc(lines, sizeof(LineInfo) * pic.height);
                memset(lines, 0, sizeof(LineInfo) * pic.height);

                calculatedLines = 0;
                outstandingLines = pic.height;
                nextline = 0;

                freeservers = nservers - 1;
                nextserver = 1;

                free(server);
                server = InitServers(nservers);

                CommRecvCancel();
                CommRecvNew();
            break;

            case C_TEST:
                state = C_LAUNCH;
                if (!CommRecvTestMaster())
                    break;

                if (result[0] == RESP_ERR_NOMEM)
                {
                    // Since this server's line number entry is not reset
                    // it is never asked again.
                    Debug(DEBUG_GEN(0), "Server %d out of memory\n", CommRecvSource());
                    CommRecvNew();
                    break;
                }

                if (result[0] == RESP_DATA && result[1] == imageno)
                {
                    int s = CommRecvSource();
                    int l = (int) result[2];        // line number

                    Debug(9, "Received lines %d-%d from node %d\n", l,
                        l + server[s].nlines - 1, s);

                    if (s >= 0 && s < nservers && server[s].line >= 0)
                    {
                        int i = 0;

                        server[s].line = -1;
                        ++freeservers;

                        if (result[4] >= TIME_EPSILON)
                        {
                            // If seconds larger than 0, calculate speed
                            // as iterations / second.
                            totalspeed -= server[s].speed;
                            server[s].speed = result[3] / result[4];
                            totalspeed += server[s].speed;
                            Debug(8, "Node %d iter %9.2f time %9.2f speed %9.2f\n",
                                s, result[3], result[4], server[s].speed);
                        }

                        while (--server[s].nlines >= 0)
                        {
                            if (lines[l + i].iterations == 0)
                            {
                                lines[l + i].iterations = result[3] / (server[s].nlines + i + 1);
                                if (l + i >= pic.height)
                                    Debug(0, "Illegal line %d\n", l + i);
                                else
                                    DisplayLine(&result[i * pic.width], l + i, pic.width);
                                --outstandingLines;
                                if (outstandingLines > 0 && l + i == calculatedLines)
                                {
                                    while (lines[calculatedLines].iterations != 0)
                                        if (++calculatedLines == pic.height)
                                            break;
                                    pic.row = calculatedLines;
                                }
                            }
                        ++i;
                        }
                    }
                }

                CommRecvNext();

                if (outstandingLines == 0)
                {
                    int i;
                    double totMFlops = 0.0;

                    if (gettimeofday(&tv, NULL) > -1
                            && tvStart.tv_sec >= 0 && tvStart.tv_usec >= 0)
                    {
                        double seconds = tv.tv_sec - tvStart.tv_sec
                            + (tv.tv_usec - tvStart.tv_usec) / 1.0e6;
                        Debug(0, "Total time: %.2f seconds\n", seconds);
                    }

                    // This calculates only the speed of the last lines.
                    for (i = 1; i < nservers; i++)
                    {
                        totMFlops += MFLOP_FACTOR * server[i].speed;
                        Debug(0, "Node %4d  %9.2f MFlops\n", i, MFLOP_FACTOR * server[i].speed);
                    }

                    Debug(0, "---------------------------\n");
                    Debug(0, "Total      %9.2f MFlops\n", totMFlops);

                    XtVaSetValues(mscreen.toplevel, XtNtitle, PROGNAME, 0);
                    state = C_DISPATCH;
                }
            break;

            case C_LAUNCH:
                state = C_DISPATCH;

                if (freeservers == 0)
                    break;

                if (calculatedLines >= pic.height)
                {
                    Debug(0, "All lines calculated but still some (%d) outstanding???\n",
                        outstandingLines);
                }

                // Search next line for which no result is calculated yet.
                // Do this in round-robin manner, and adjust for server speed.
                // This loop must succeed because freeservers > 0.
                while (server[nextserver].line != -1)
                {
                    if (++nextserver > nservers)
                        nextserver = 1;
                }

                int oldline = nextline;

                // iterations is 0 as long as the line is not calculated yet.
                // We can use this for the loop test.
                while (lines[nextline].iterations != 0)
                {
                    if (++nextline >= pic.height)
                        nextline = calculatedLines;
                    if (nextline == oldline)
                    {
                        Debug(0, "Line calculation wrong, finishing\n");
                        outstandingLines = 0;
                        break;
                    }
                }

                if (nextline >= pic.height)
                    break;

                server[nextserver].line = nextline;
                server[nextserver].nlines = 1;
                --freeservers;

                command = CommGetCommand();
                memcpy(&command->pic, &pic, sizeof pic);
                command->cmd = CMD_START;
                command->imageno = imageno;
                command->lineno = nextline;

                if (totalspeed > 0.0)
                {
                    int tmp = (int) (server[nextserver].speed / totalspeed
                        * pic.height);
                    if (tmp > 0)
                    {
                        if (tmp > (pic.height / MAX_LINE_FACTOR))
                            tmp = (pic.height / MAX_LINE_FACTOR);
                        if (tmp > pic.height - nextline)
                            tmp = pic.height - nextline;
                        // Verify that the lines are not calculated yet.
                        while (lines[nextline + tmp - 1].iterations != 0)
                            --tmp;
                        server[nextserver].nlines = tmp;
                    }
                }

                command->nlines = server[nextserver].nlines;

                Debug(9, "Lines %d-%d sent to node %d\n", nextline,
                    nextline + server[nextserver].nlines - 1, nextserver);
                CommSend(nextserver);

                nextline += server[nextserver].nlines;
                if (nextline >= pic.height)
                    nextline = calculatedLines;
            break;

        } /* switch */

    } /* while */

    return 0;               /* Never reached */

} /* main */
