/*
 *  This file implements a semaphore API that is easy to understand and free of
 *  race conditions.
 *  Adapted from code in W.R. Stevens' "Unix network programming".
 *
 *  HISTORY
 *  Date        Author  Remarks
 *  1999-03-07  hjb     Created
 */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <errno.h>
#include "semlib.h"

static const int BIGCOUNT = 10000;      /* Initial value of process counter */

static struct sembuf op_lock[2] =
{
    { 2, 0, 0,       },     /* Wait for [2] (lock) to equal 0 */
    { 2, 1, SEM_UNDO }      /* Undo to release the lock if process exits
                               before explicitly unlocking. */
};

static struct sembuf op_endcreate[2] =
{
    { 1, -1, SEM_UNDO },    /* Decrement [1] (proc counter) with undo on
                               exit. Undo to adjust proc counter if process
                               exits before explicitly calling sem_close. */
    { 2, -1, SEM_UNDO }     /* then decrement [2] (lock) back to 0 */
};

static struct sembuf op_open[1] =
{
    { 1, -1, SEM_UNDO }     /* Decrement [1] (proc counter) with undo on exit */
};

static struct sembuf op_close[3] =
{
    { 2, 0, 0        },     /* wait for [2] (lock) to equal 0 */
    { 2, 1, SEM_UNDO },     /* then increment [2] to 1 - this locks it */
    { 1, 1, SEM_UNDO }      /* then increment [1] (proc counter) */
};

static struct sembuf op_unlock[1] =
{
    { 2, -1, SEM_UNDO }     /* decrement [2] (lock) back to 0 */
};

static struct sembuf op_op[1] =
{
    { 0, 99, SEM_UNDO }     /* Decrement or increment [0] with undo on exit.
                               The 99 is set to the actual amount to add or
                               subtract (positive or negative). */
};

union
{
    int             val;
    struct semid_ds *buf;
    ushort          *array;
} semctl_arg;

int sem_create(key_t key, int initval)
{
    int     id;
    int     semval;

    if (key == IPC_PRIVATE)
        return -1;          /* not intended for private semaphores */
    else if (key == (key_t) -1)
        return -1;          /* probably an ftok() error by caller */

again:
    if ((id = semget(key, 3, 0666 | IPC_CREAT)) < 0)
        return -1;          /* permission problem or tables full */

    if (semop(id, op_lock, 2) < 0)
    {
        if (errno == EINVAL)
            goto again;
        return -1;          /* cannot lock */
    }

    semctl_arg.val = 0;
    if ((semval = semctl(id, 1, GETVAL, semctl_arg)) < 0)
        return -1;

    if (semval == 0)
    {
        semctl_arg.val = initval;
        if (semctl(id, 0, SETVAL, semctl_arg) < 0)
            return -1;

        semctl_arg.val = BIGCOUNT;
        if (semctl(id, 1, SETVAL, semctl_arg) < 0)
            return -1;
    }

    /* Decrement the process counter and then release the lock. */
    if (semop(id, op_endcreate, 2) < 0)
        return -1;

    return id;
}

int sem_open(key_t key)
{
    int     id;

    if (key == IPC_PRIVATE)
        return -1;          /* not intended for private semaphores */

    if ((id = semget(key, 3, 0)) < 0)
        return -1;          /* doesn't exist, or tables full */
    else if (key == (key_t) -1)
        return -1;          /* probably an ftok() error by caller */

    /* Decrement the process counter. We don't need a lock to do this. */
    if (semop(id, op_open, 1) < 0)
        return -1;

    return id;
}

void sem_rm(int id)
{
    semctl_arg.val = 0;
    semctl(id, 0, IPC_RMID, semctl_arg);
}

void sem_close(int id)
{
    int semval;

    if (semop(id, op_close, 3) < 0)
        return;

    semctl_arg.val = 0;
    if ((semval = semctl(id, 1, GETVAL, semctl_arg)) < 0)
        return;

    if (semval == BIGCOUNT)
        sem_rm(id);
    else
        semop(id, op_unlock, 1);
}

void sem_wait(int id)
{
    sem_op(id, -1);
}

void sem_signal(int id)
{
    sem_op(id, 1);
}

void sem_op(int id, int value)
{
    op_op[0].sem_op = value;
    if (value == 0)
        return;

    semop(id, op_op, 1);
}
