#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <Xlib.h>
#include <X11/cursorfont.h>
#include "dialog.h"
#include "mandel.h"


Image	readpic;
static long	headerlen;	/* file offset to start of image data */
static int	maxcolor;	/* max RGB values in file (max 255) */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

unsigned long NearestPixel (XColor *colormap, int r, int g, int b)
{
  int	i;
  int	index;
  long	diff, best;

  best = 0xffffff;

  for (i = 0; i < 256; i++)
  {
    diff = abs (colormap [i].red - r) + abs (colormap [i].blue - b)
      + abs (colormap [i].green - g);
    if (diff < best)
    {
      best = diff;
      index = i;
      if (diff == 0)		/* speedup */
        return i;
    }
  }
  return index;
} /* NearestPixel */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

Boolean WriteImage (const char *filename)
{
  FILE		*f;
  int		row, col;
  unsigned long	color;
  XColor	colormap [256];			/* palette for depth 8 */

  f = fopen (filename, "wb");
  if (f == 0)
  {
    Debug (0, "Could not open file %s\n", filename);
    return False;
  }

  fprintf (f, "P6\n");
  /* We write the image parameters as comments directly after the header.
     Our comments are marked with ' so anybody can add her own comments
     as long as these do not begin with '. */
  fprintf (f, "#  Real coordinates begin : end\n");
  fprintf (f, "#' %20.18f %20.18f\n", pic.re1, pic.re2);
  fprintf (f, "#  Imaginary coordinates begin : end\n");
  fprintf (f, "#' %20.18f %20.18f\n", pic.im1, pic.im2);
  fprintf (f, "#  Perturbation real : imaginary\n");
  fprintf (f, "#' %20.18f %20.18f\n", pic.reP, pic.imP);
  fprintf (f, "#  Julia point real : imaginary\n");
  fprintf (f, "#' %20.18f %20.18f\n", pic.reJ, pic.imJ);
  fprintf (f, "#  Type : bailout value : max. iterations\n");
  fprintf (f, "#' %s %20.18f %lu\n", pic.julia ? "Julia" : "Mandel",
    pic.bailout, pic.maxc);
  fprintf (f, "#  Lines : completed\n");
  fprintf (f, "#' %d %d\n", pic.height, pic.row);
  /* An empty comment denotes the end of our header. So later versions
     with larger headers stay compatible. */
  fprintf (f, "#\n");

  fprintf (f, "%d %d\n", pic.width, pic.row);
  fprintf (f, "255\n");

  if (mscreen.depth == 8)
    GetColormap (colormap);

  /* Now follows the image in binary form. Only the completed lines
     are written. */
  for (row = 0; row < pic.row; row++)
    for (col = 0; col < pic.width; col++)
    {
      color = XGetPixel (mscreen.image, col, row);
      switch (mscreen.depth)
      {
        case 8:
	  fprintf (f, "%c%c%c", colormap [color].red,
	    colormap [color].green, colormap [color].blue);
	break;
        case 15:
	  fprintf (f, "%c%c%c", (char) ((color >> 7) & 0xf1),
	    (char) ((color >> 2) & 0xf1), (char) ((color & 31) << 3));
        break;
        case 16:
	  if (mscreen.weight565)
	    fprintf (f, "%c%c%c", (char) ((color >> 8) & 0xf1),
	      (char) ((color >> 3) & 0xf3), (char) ((color & 31) << 3));
	  else
	    fprintf (f, "%c%c%c", (char) ((color >> 8) & 0xf3),
	      (char) ((color >> 2) & 0xf1), (char) ((color & 31) << 3));
        break;
        case 24:
	  fprintf (f, "%c%c%c", char (color >> 16), char (color >> 8),
	    char (color));
        break;
	default:
	  Debug (0, "Invalid depth in WriteImage\n");
      } /* switch */
      if (errno)
      {
        Debug (0, "Could not write to file %s\n", filename);
	return False;
      }
    } /* for row, col */

  return fclose (f) == 0;
} /* WriteImage */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

Boolean OpenImage (const char *filename)
{
  FILE	*f;
  char	*buffer;		/* buffer for one line */
  char	imtype [100];		/* buffer for image type */
  int	bsize = 100;		/* length of buffer, may increase */
  int	len;
  int	numitems;

  enum
  {
    HS_MAGIC,
    HS_WIDTH,
    HS_HEIGHT,
    HS_MAXCOLOR
  } hstate = HS_MAGIC;	/* which component of the header is expected next */

  enum
  {
    MS_REAL,
    MS_IMAG,
    MS_PERT,
    MS_JULIA,
    MS_OPT,
    MS_LINE,
    MS_READY
  } mstate = MS_REAL;

  f = fopen (filename, "rb");
  if (f == 0)
    return False;
  memset (&readpic, 0, sizeof (readpic));
  buffer = (char *) malloc (bsize);
  if (buffer == 0)
  {
    Debug (0, "Out of memory\n");
    fclose (f);
    return False;
  }

  headerlen = 0;
  do
  {
    if (fgets (buffer, bsize, f) == 0)
    {
      fclose (f);
      free (buffer);
      return False;
    }
    len = strlen (buffer);
    while (len > 0 && buffer [len - 1] != '\n')
    {
      /* Allocate more memory and read rest of line */
      bsize += 200;
      if (realloc (buffer, bsize) == 0)
      {
        Debug (0, "Out of memory\n");
        fclose (f);
        free (buffer);
        return False;
      }
      if (fgets (&buffer [len], 200, f) == 0)
      {
        fclose (f);
        free (buffer);
        return False;
      }
      len = strlen (buffer);
    } /* while */

    if (len == 0)
      continue;
    if (hstate == HS_MAGIC)
      if (buffer [0] == 'P' && buffer [1] == '6')
      {
        hstate = HS_WIDTH;
	continue;
      }
      else
      {
        fclose (f);
        free (buffer);
        return False;
      }

    if (buffer [0] == '#')
    {
      if (buffer [1] == '\'')
      switch (mstate)
      {
        case MS_REAL:
	  if (sscanf (&buffer [2], "%lf %lf", &readpic.re1, &readpic.re2) == 2)
	    mstate = MS_IMAG;
	  else
            {
              fclose (f);
              free (buffer);
              return False;
            }
	break;

        case MS_IMAG:
	  if (sscanf (&buffer [2], "%lf %lf", &readpic.im1, &readpic.im2) == 2)
	    mstate = MS_PERT;
	  else
            {
              fclose (f);
              free (buffer);
              return False;
            }
	break;

        case MS_PERT:
	  if (sscanf (&buffer [2], "%lf %lf", &readpic.reP, &readpic.imP) == 2)
	    mstate = MS_JULIA;
	  else
            {
              fclose (f);
              free (buffer);
              return False;
            }
	break;

        case MS_JULIA:
	  if (sscanf (&buffer [2], "%lf %lf", &readpic.reJ, &readpic.imJ) == 2)
	    mstate = MS_OPT;
	  else
            {
              fclose (f);
              free (buffer);
              return False;
            }
	break;

/* FIXME: range checks */

        case MS_OPT:
	  if (sscanf (&buffer [2], "%20s %lf %lu", imtype, &readpic.bailout,
	      &readpic.maxc) == 3)
	  {
	    mstate = MS_LINE;
	    if (strcmp (imtype, "Mandel") == 0)
	      readpic.julia = False;
	    else if (strcmp (imtype, "Julia") == 0)
	      readpic.julia = True;
	    else
            {
              fclose (f);
              free (buffer);
              return False;
            }
	  }
	  else
          {
            fclose (f);
            free (buffer);
            return False;
          }
	break;

        case MS_LINE:
	  if (sscanf (&buffer [2], "%d %d", &readpic.height, &readpic.row) == 2)
	    mstate = MS_READY;
	  else
          {
            fclose (f);
            free (buffer);
            return False;
          }
	break;

        case MS_READY:
          fclose (f);
          free (buffer);
          return False;
	break;

        default:
	  ;
      } /* switch */
    }
    else
    {
      switch (hstate)
      {
        case HS_WIDTH:
	  numitems = sscanf (buffer, "%u %u %u", &readpic.width, &readpic.row,
	    &maxcolor);
	  switch (numitems)
	  {
	    case EOF:
	    case 0:
	      Debug (0, "Error while reading header of %s\n", filename);
	    break;
	    case 1:
	      hstate = HS_HEIGHT;
	    break;
	    case 2:
	      hstate = HS_MAXCOLOR;
	    break;
	    case 3:
	      if (mstate == MS_READY)
                headerlen = ftell (f);
	      else
              {
                fclose (f);
                free (buffer);
                return False;
              }
	    break;
	    default:
	      Debug (0, "Unexpected switch value in ReadImage\n");
	      fclose (f);
	      return False;
	  }
	break;

        case HS_HEIGHT:
	  numitems = sscanf (buffer, "%u %u", &readpic.row, &maxcolor);
	  switch (numitems)
	  {
	    case EOF:
	    case 0:
	      Debug (0, "Error while reading header of %s\n", filename);
	    break;
	    case 1:
	      hstate = HS_MAXCOLOR;
	    break;
	    case 2:
	      if (mstate == MS_READY)
                headerlen = ftell (f);
	      else
              {
                fclose (f);
                free (buffer);
                return False;
              }
	    break;
	    default:
	      Debug (0, "Unexpected switch value in ReadImage\n");
	      fclose (f);
	      return False;
	  }
	break;

        case HS_MAXCOLOR:
	  numitems = sscanf (buffer, "%u", &maxcolor);
	  switch (numitems)
	  {
	    case EOF:
	    case 0:
	      Debug (0, "Error while reading header of %s\n", filename);
	    break;
	    case 1:
	      if (mstate == MS_READY)
                headerlen = ftell (f);
	      else
              {
                fclose (f);
                free (buffer);
                return False;
              }
	    break;
	    default:
	      Debug (0, "Unexpected switch value in ReadImage\n");
	      fclose (f);
	      return False;
	  }
	break;

        default:
	  ;
      } /* switch */
    }
  } while (headerlen == 0);

  fclose (f);
  free (buffer);
  pic = readpic;
  return True;
} /* OpenImage */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

Boolean ReadImage (const char *filename)
{
  FILE		*f;
  int		row, col;
  unsigned long	color;
  int		r, g, b;
  XColor	colormap [256];		/* palette for depth 8 */
  Cursor	cursor;
  int		colorscale;
  int		coloroffset;

  f = fopen (filename, "rb");
  if (f == 0)
    return False;
  if (fseek (f, headerlen, SEEK_SET) == -1)
  {
    fclose (f);
    return False;
  }

  NewSize ();
  if (mscreen.depth == 8)
    GetColormap (colormap);

  /*
   * RGB values from the file get multiplied by colorscale to lie in the
   * 0..255 range. If colorscale must be rounded down, addition of
   * coloroffset adjusts the range for brighter colors. Otherwise
   * coloroffset is 0.
   */
  colorscale = 256 / (maxcolor + 1);
  coloroffset = 256 - colorscale * (maxcolor + 1);

  cursor = XCreateFontCursor (mscreen.display, XC_watch);
  XDefineCursor (mscreen.display, XtWindow (mscreen.toplevel), cursor);
  XFlush (mscreen.display);

  for (row = 0; row < pic.row; row++)
    for (col = 0; col < pic.width; col++)
    {
      r = fgetc (f) * colorscale + coloroffset;
      g = fgetc (f) * colorscale + coloroffset;
      b = fgetc (f) * colorscale + coloroffset;
      if (errno)
      {
        XDefineCursor (mscreen.display, XtWindow (mscreen.toplevel), None);
        XFreeCursor (mscreen.display, cursor);
        Debug (0, "Could not read to file %s\n", filename);
	return False;
      }
      switch (mscreen.depth)
      {
        case 8:
	  r= r << 8;
	  g = g << 8;
	  b = b << 8;
	  color = NearestPixel (colormap, r, g, b);
	break;
        case 15:
	  color = ((r & 0xf1) << 7) + ((g & 0xf1) << 2) + (b >> 3);
        break;
        case 16:
	  if (mscreen.weight565)
	    color = ((r & 0xf1) << 8) + ((g & 0xf3) << 3) + (b >> 3);
	  else
	    color = ((r & 0xf3) << 8) + ((g & 0xf1) << 2) + (b >> 3);
        break;
        case 24:
	  color = (r << 16) + (g << 8) + b;
        break;
	default:
	  Debug (0, "Invalid depth in WriteImage\n");
      } /* switch */
      XPutPixel (mscreen.image, col, row, color);
    } /* for row, col */

  fclose (f);
  XDefineCursor (mscreen.display, XtWindow (mscreen.toplevel), None);
  XFreeCursor (mscreen.display, cursor);
  newPicture = True;
  return True;
} /* ReadImage */




/*---------------------------------------------------------------------------
 *
 *
 *---------------------------------------------------------------------------*/

Boolean GetImage (const char *filename)
{
  if (OpenImage (filename))
  {
    if (ReadImage (filename))
      return True;
    else
      Debug (0, "Could not read image %s\n", "");
  }
  else
    Debug (0, "Could not open image %s\n", "");
  return False;
} /* GetImage */

